/* ncdu - NCurses Disk Usage 
  
  Copyright (c) 2007 Yoran Heling

  Permission is hereby granted, free of charge, to any person obtaining
  a copy of this software and associated documentation files (the
  "Software"), to deal in the Software without restriction, including
  without limitation the rights to use, copy, modify, merge, publish,
  distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to
  the following conditions:
  
  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include "ncdu.h"

struct dir *bcur, *bsel;
int helpwin;

int cmp(struct dir *x, struct dir *y) {
  struct dir *a, *b;
  int r = 0;
  if(bflags & BF_DESC) {
    a = y; b = x;
  } else {
    b = y; a = x;
  }
  if(!(bflags & BF_NDIRF) && y->flags & FF_DIR && !(x->flags & FF_DIR))
    r = 1;
  else if(!(bflags & BF_NDIRF) && !(y->flags & FF_DIR) && x->flags & FF_DIR)
    r = -1;
  else if(bflags & BF_NAME)
    r = strcmp(a->name, b->name);
  else if(bflags & BF_FILES)
    r = (a->files - b->files);
  else if(bflags & BF_DIRS)
    r = (a->dirs - b->dirs);
  if(r == 0)
    r = a->size > b->size ? 1 : (a->size == b->size ? 0 : -1);
  return(r);
}

/* Mergesort algorithm, many thanks to
   http://www.chiark.greenend.org.uk/~sgtatham/algorithms/listsort.html
*/
struct dir *sortFiles(struct dir *list) {
  struct dir *p, *q, *e, *tail;
  int insize, nmerges, psize, qsize, i;

  while(list->prev != NULL)
    list = list->prev;
  insize = 1;
  while(1) {
    p = list;
    list = NULL;
    tail = NULL;
    nmerges = 0;
    while(p) {
      nmerges++;
      q = p;
      psize = 0;
      for(i=0; i<insize; i++) {
        psize++;
        q = q->next;
        if(!q) break;
      }
      qsize = insize;
      while(psize > 0 || (qsize > 0 && q)) {
        if(psize == 0) {
          e = q; q = q->next; qsize--;
        } else if(qsize == 0 || !q) {
          e = p; p = p->next; psize--;
        } else if(cmp(p,q) <= 0) {
          e = p; p = p->next; psize--;
        } else {
          e = q; q = q->next; qsize--;
        }
        if(tail) tail->next = e;
        else     list = e;
        e->prev = tail;
        tail = e;
      }
      p = q;
    }
    tail->next = NULL;
    if(nmerges <= 1)
      return list;
    insize *= 2;
  }
}

int helpScreen(void) {
  WINDOW *hlp;
  int ch=0, x, y;

  hlp = newwin(14, 60, winrows/2-6, wincols/2-30);
  curs_set(0);
  if(!helpwin) helpwin = 1;
  do {
    if(ch == ERR)
      continue;
    if(ch && ch != '1' && ch != '2' && ch != '3')
      break;
    if(ch)
      helpwin = ch - 48;

    werase(hlp);
    box(hlp, 0, 0);
    wattron(hlp, A_BOLD);
    mvwaddstr(hlp, 0, 4, "ncdu help");
    wattroff(hlp, A_BOLD);
    mvwaddstr(hlp, 12, 32, "Press any key to continue");
    switch(helpwin) {
      case 1:
        wattron(hlp, A_BOLD);
        mvwaddstr(hlp, 1, 30, "1:Keys");
        wattroff(hlp, A_BOLD);
        mvwaddstr(hlp, 1, 39, "2:Format");
        mvwaddstr(hlp, 1, 50, "3:About");
        wattron(hlp, A_BOLD);
        mvwaddstr(hlp, 3,  7, "up/down");
        mvwaddstr(hlp, 4,  3, "right/enter");
        mvwaddstr(hlp, 5, 10, "left");
        mvwaddstr(hlp, 6, 13, "n");
        mvwaddstr(hlp, 7, 13, "s");
        mvwaddstr(hlp, 8, 13, "p");
        mvwaddstr(hlp, 9, 13, "h");
        mvwaddstr(hlp,10, 13, "q");
        wattroff(hlp, A_BOLD);
        mvwaddstr(hlp, 3, 16, "Cycle through the items");
        mvwaddstr(hlp, 4, 16, "Open directory");
        mvwaddstr(hlp, 5, 16, "Previous directory");
        mvwaddstr(hlp, 6, 16, "Sort by name (ascending/descending)");
        mvwaddstr(hlp, 7, 16, "Sort by size (ascending/descending)");
        mvwaddstr(hlp, 8, 16, "Toggle between powers of 1000 and 1024");
        mvwaddstr(hlp, 9, 16, "Show/hide hidden files");
        mvwaddstr(hlp,10, 16, "Quit ncdu");
        break;
      case 2:
        mvwaddstr(hlp, 1, 30, "1:Keys");
        wattron(hlp, A_BOLD);
        mvwaddstr(hlp, 1, 39, "2:Format");
        wattroff(hlp, A_BOLD);
        mvwaddstr(hlp, 1, 50, "3:About");
        wattron(hlp, A_BOLD);
        mvwaddstr(hlp, 3, 3, "X  [size]  [file or directory]");
        wattroff(hlp, A_BOLD);
        mvwaddstr(hlp, 5, 4, "The X is only present in the following cases:");
        wattron(hlp, A_BOLD);
        mvwaddstr(hlp, 6, 4, "!");
        mvwaddstr(hlp, 7, 4, ".");
        mvwaddstr(hlp, 8, 4, ">");
        mvwaddstr(hlp, 9, 4, "@");
        mvwaddstr(hlp,10, 4, "e");
        wattroff(hlp, A_BOLD);
        mvwaddstr(hlp, 6, 7, "An error occured while reading this directory");
        mvwaddstr(hlp, 7, 7, "An error occured while reading a subdirectory");
        mvwaddstr(hlp, 8, 7, "Directory was on an other filesystem");
        mvwaddstr(hlp, 9, 7, "This is not a file nor a dir (symlink, socket, ...)");
        mvwaddstr(hlp,10, 7, "Empty directory");
        break;
      case 3:
        /* Indeed, too much spare time */
        mvwaddstr(hlp, 1, 30, "1:Keys");
        mvwaddstr(hlp, 1, 39, "2:Format");
        wattron(hlp, A_BOLD);
        mvwaddstr(hlp, 1, 50, "3:About");
        wattroff(hlp, A_BOLD);
        wattron(hlp, A_REVERSE);
        x=12;y=3;
        /* N */
        mvwaddstr(hlp, y+0, x+0, "      ");
        mvwaddstr(hlp, y+1, x+0, "  ");
        mvwaddstr(hlp, y+2, x+0, "  ");
        mvwaddstr(hlp, y+3, x+0, "  ");
        mvwaddstr(hlp, y+4, x+0, "  ");
        mvwaddstr(hlp, y+1, x+4, "  ");
        mvwaddstr(hlp, y+2, x+4, "  ");
        mvwaddstr(hlp, y+3, x+4, "  ");
        mvwaddstr(hlp, y+4, x+4, "  ");
        /* C */
        mvwaddstr(hlp, y+0, x+8, "     ");
        mvwaddstr(hlp, y+1, x+8, "  ");
        mvwaddstr(hlp, y+2, x+8, "  ");
        mvwaddstr(hlp, y+3, x+8, "  ");
        mvwaddstr(hlp, y+4, x+8, "     ");
        /* D */
        mvwaddstr(hlp, y+0, x+19, "  ");
        mvwaddstr(hlp, y+1, x+19, "  ");
        mvwaddstr(hlp, y+2, x+15, "      ");
        mvwaddstr(hlp, y+3, x+15, "  ");
        mvwaddstr(hlp, y+3, x+19, "  ");
        mvwaddstr(hlp, y+4, x+15, "      ");
        /* U */
        mvwaddstr(hlp, y+0, x+23, "  ");
        mvwaddstr(hlp, y+1, x+23, "  ");
        mvwaddstr(hlp, y+2, x+23, "  ");
        mvwaddstr(hlp, y+3, x+23, "  ");
        mvwaddstr(hlp, y+0, x+27, "  ");
        mvwaddstr(hlp, y+1, x+27, "  ");
        mvwaddstr(hlp, y+2, x+27, "  ");
        mvwaddstr(hlp, y+3, x+27, "  ");
        mvwaddstr(hlp, y+4, x+23, "      ");
        wattroff(hlp, A_REVERSE);
        mvwaddstr(hlp, y+0, x+30, "NCurses");
        mvwaddstr(hlp, y+1, x+30, "Disk");
        mvwaddstr(hlp, y+2, x+30, "Usage");
        mvwprintw(hlp, y+4, x+30, "%s", PACKAGE_VERSION);

        mvwaddstr(hlp, 9,  7, "Written by Yoran Heling <projects@yorhel.nl>");
        mvwaddstr(hlp,10, 16, "http://dev.yorhel.nl/ncdu/");
        break;
    }
    wrefresh(hlp);
  } while((ch = getch()));
  delwin(hlp);
  touchwin(stdscr);
  refresh();
  if(ch == KEY_RESIZE) {
    getmaxyx(stdscr, winrows, wincols);
    return(1);
  }
  helpwin = 0;
  return(0);
}

void browseDir(void) {
  ITEM **items;
  char **itrows, tmp[PATH_MAX];
  MENU *menu;
  struct dir *d;
  int i, fls, ch, bf, s;

  erase();
  attron(A_REVERSE);
  for(i=0; i<wincols;i++) {
    mvaddch(0, i, ' ');
    mvaddch(winrows-1, i, ' ');
  }
  mvprintw(0,0,"%s %s ~ Use the arrow keys to navigate, press ? for help", PACKAGE_NAME, PACKAGE_VERSION);
  mvprintw(winrows-1, 0, " Total size: %s   Files: %-6d   Dirs: %-6d",
      cropsize(bcur->parent->size), bcur->parent->files, bcur->parent->dirs);
  attroff(A_REVERSE);
  for(i=wincols; i>=0; i--)
    mvaddch(1,i,'-');

  bcur = sortFiles(bcur);
  fls = 1; d = bcur;
  while(d != NULL) {
    if(!(bflags & BF_HIDE) || d->name[0] != '.')
      fls++;
    d = d->next;
  }

  d = bcur;
  while(d->parent != NULL) {
    d->parent->sub = d;
    d = d->parent;
  }
  tmp[0] = '\0';
  while(d->parent != bcur->parent) {
    if(d->parent != NULL && d->parent->name[strlen(d->parent->name)-1] != '/') strcat(tmp, "/");
    strcat(tmp, d->name);
    d = d->sub;
  }
  mvaddstr(1, 3, cropdir(tmp, wincols-5));

  items = calloc(fls+2, sizeof(ITEM *));
  itrows = malloc((fls+1) * sizeof(char *));
  itrows[0] = malloc((fls+1) * wincols * sizeof(char));
  for(i=1; i<fls; i++)
    itrows[i] = itrows[0] + i*wincols;

  i=0; s=-1;
  sprintf(tmp, "%%c %%7s  %%c%%-%ds", wincols-12);
  if(bcur->parent->parent != NULL) {
    items[i] = new_item("           /..", 0);
    set_item_userptr(items[i], NULL);
    i++;
  }
  d = bcur; 
  while(d != NULL) {
    if(bflags & BF_HIDE && d->name[0] == '.') {
      d = d->next;
      continue;
    }
    sprintf(itrows[i], tmp, 
        d->flags & FF_ERR ? '!' : d->flags & FF_SERR ? '.' : d->flags & FF_OTHFS ? '>' :
          d->flags & FF_OTHER ? '@' : d->flags & FF_DIR && d->sub == NULL ? 'e' : ' ',
        cropsize(d->size), d->flags & FF_DIR ? '/' : ' ', cropdir(d->name, wincols-12));
    items[i] = new_item(itrows[i], 0);
    set_item_userptr(items[i], d);
    if(bsel == d) s = i;
    d = d->next;
    i++;
  }
  items[i] = NULL;
  menu = new_menu(items);
  set_menu_format(menu, winrows-3, 1);
  set_menu_sub(menu, derwin(stdscr, winrows-3, wincols, 2, 0));
  set_menu_mark(menu, "");
  post_menu(menu);
  if(s >= 0) set_current_item(menu, items[s]);
  
  d = bcur; bf = bflags;
  refresh();
  
  if(helpwin > 0 && helpScreen() == 1) goto endbrowse;
  while((ch = getch())) {
    switch(ch) {
      case KEY_UP:    menu_driver(menu, REQ_UP_ITEM);    break;
      case KEY_DOWN:  menu_driver(menu, REQ_DOWN_ITEM);  break;
      case KEY_HOME:  menu_driver(menu, REQ_FIRST_ITEM); break;
      case KEY_LL: 
      case KEY_END:   menu_driver(menu, REQ_LAST_ITEM);  break;
      case KEY_PPAGE: menu_driver(menu, REQ_SCR_UPAGE);  break;
      case KEY_NPAGE: menu_driver(menu, REQ_SCR_DPAGE);  break;
      case '?':
        if(helpScreen() == 1)
          goto endbrowse;
        break;
      case 'q':
        bcur = NULL;
        goto endbrowse;
      case 10: case KEY_RIGHT:
        bcur = item_userptr(current_item(menu));
        if(bcur == NULL || bcur->sub != NULL) {
          bcur = bcur == NULL ? d->parent : bcur->sub;
          goto endbrowse;
        } else
          bcur = d;
        break;
      case KEY_RESIZE:
        getmaxyx(stdscr, winrows, wincols);
        goto endbrowse;
      case KEY_LEFT:
        if(bcur->parent->parent != NULL)
          bcur = bcur->parent;
        else
          bcur = d;
        goto endbrowse;
      case 'n':
        if(bflags & BF_NAME) {
          if(bflags & BF_DESC) bflags -= BF_DESC;
          else bflags |= BF_DESC;
        } else
          bflags = (bflags & BF_HIDE) + (bflags & BF_NDIRF) + BF_NAME;
        goto endbrowse;
      case 's':
        if(bflags & BF_SIZE) {
          if(bflags & BF_DESC) bflags -= BF_DESC;
          else bflags |= BF_DESC;
        } else
          bflags = (bflags & BF_HIDE) + (bflags & BF_NDIRF) + BF_SIZE;
        goto endbrowse;
      case 'p':
        if(sflags & SF_SI)
          sflags = (sflags & SF_SMFS) + (sflags & SF_AS);
        else
          sflags |= SF_SI;
        goto endbrowse;
      case 'h':
        if(bflags & BF_HIDE)
          bflags -= BF_HIDE;
        else
          bflags |= BF_HIDE;
        goto endbrowse;
    }
    refresh();
  }
  endbrowse:
/* Setting bsel will introduce some bugs on large lists
   if(bcur == d)
    bsel = item_userptr(current_item(menu));
  else */
    bsel = NULL;
  unpost_menu(menu);
  for(i=fls-3;i>=0;i--)
    free_item(items[i]);
  free(items);
  free(itrows[0]);
  free(itrows);
  erase();
}

void showBrowser(void) {
  bcur = dat.sub;
  bsel = NULL;
  helpwin = 0;
  while(bcur != NULL)
    browseDir();
}
